#include "multigrid.h"
#include "funcoes.h"

/* Metodos Privados */
void Multigrid::Relaxar_Av(unsigned int num_iteracoes, double t){
	unsigned int i,j,k,l,m; /*facilita paralelizao*/

   	unsigned int nx = Grids[0]->get_nx();
    unsigned int ny = Grids[0]->get_ny();

    Matriz<par>& x = Grids[0]->X;
    Matriz<double>& v = Grids[0]->v;
    Matriz<double>& v_old = Grids[0]->v_old;

    double termo_forcante;
    
	/*Red-Black Gauss-Seidel*/
	for (i=0 ; i < num_iteracoes ; i++){		
		for (j=1 ; j < nx-1 ; j=j+2){
			for (k=1 ; k < ny-1 ; k=k+2){
				termo_forcante = func_termo_forcante(x(j,k).x , x(j,k).y , t+ht); //calcula termo forcante para tempo futuro
				v(j,k) = (beta*( v(j-1,k) + v(j+1,k) ) + beta*alpha*(v(j,k-1) + v(j,k+1)) + hx2*v_old(j,k) + ht*hx2*termo_forcante)/gama;
			}
		}
	
		for (l=2 ; l < nx-1 ; l=l+2){
			for (m=2 ; m < ny-1 ; m=m+2){
				termo_forcante = func_termo_forcante(x(l,m).x , x(l,m).y , t+ht); //calcula termo forcante para tempo futuro
				v(l,m) = (beta*( v(l-1,m) + v(l+1,m) )+ beta*alpha*(v(l,m-1) + v(l,m+1)) +hx2*v_old(l,m) + ht*hx2*termo_forcante)/gama;
			}
		}
		
		for (j=1 ; j < nx-1 ; j=j+2){
			for (k=2 ; k < ny-1 ; k=k+2){
				termo_forcante = func_termo_forcante(x(j,k).x , x(j,k).y , t+ht); //calcula termo forcante para tempo futuro
				v(j,k) = (beta*( v(j-1,k) + v(j+1,k) )+ beta*alpha*(v(j,k-1) + v(j,k+1))  +v_old(j,k)*hx2 + ht*hx2*termo_forcante)/gama;
			}
		}
		
		for (l=2 ; l < nx-1 ; l=l+2){
			for (m=1 ; m < ny-1 ; m=m+2){
				termo_forcante = func_termo_forcante(x(l,m).x , x(l,m).y , t+ht); //calcula termo forcante para tempo futuro
				v(l,m) = (beta*( v(l-1,m) + v(l+1,m) )+ beta*alpha*(v(l,m-1) + v(l,m+1))  + v_old(l,m)*hx2 + ht*hx2*termo_forcante)/gama;
			}
		}
	}
}

void Multigrid::Relaxar_Ac(Grid* g, unsigned int num_iteracoes){
	unsigned int i,j,k,l,m;

    unsigned  int nx = g->get_nx();
    unsigned int ny = g->get_ny();

    Matriz<double>& c = g->c;
    Matriz<double>& r = g->r;

	/*Red-Black Gauss-Seidel*/
	for (i=0 ; i < num_iteracoes ; i++){
		for (j=1 ; j < nx-1 ; j=j+2){
			for (k=1 ; k < ny-1 ; k=k+2){
				c(j,k) = (beta*( c(j-1,k) + c(j+1,k)) + beta*alpha*(c(j,k-1) + c(j,k+1) ) +r(j,k))/gama;
			}
		}
	
		for (l=2 ; l < nx-1 ; l=l+2){
			for (m=2 ; m < ny-1 ; m=m+2){
				c(l,m) = (beta*( c(l-1,m) + c(l+1,m)) + beta*alpha*(c(l,m-1) + c(l,m+1) ) +r(l,m))/gama;
			}
		}
		
		for (j=1 ; j < nx-1 ; j=j+2){
			for (k=2 ; k < ny-1 ; k=k+2){
				c(j,k) = (beta*( c(j-1,k) + c(j+1,k)) + beta*alpha*(c(j,k-1) + c(j,k+1) ) +r(j,k))/gama;
			}
		}
	

		for (l=2 ; l < nx-1 ; l=l+2){
			for (m=1 ; m < ny-1 ; m=m+2){
				c(l,m) = (beta*( c(l-1,m) + c(l+1,m)) + beta*alpha*(c(l,m-1) + c(l,m+1) ) +r(l,m))/gama;
			}
		}
	}
}

void Multigrid::calcular_residuo_fine(double t){
	int j,k;
	int nx = Grids[0]->get_nx();
    int ny = Grids[0]->get_ny();

    Matriz<par>& x = Grids[0]->X;
    Matriz<double>& r = Grids[0]->r;
    Matriz<double>& v = Grids[0]->v;
    Matriz<double>& v_old = Grids[0]->v_old;

    double termo_forcante;
	for (j=1 ; j < nx-1 ; j++){
		for (k=1 ; k < ny-1 ; k++){
			termo_forcante = func_termo_forcante(x(j,k).x , x(j,k).y , t+ht); //calcula termo forcante para tempo futuro
			r(j,k) = ht*hx2*termo_forcante + v_old(j,k)*hx2 + beta*(v(j-1,k) + v(j+1,k)) + beta*alpha*(v(j,k+1) + v(j,k-1)) -v(j,k)*gama;
		}
	}
}
void Multigrid::calcular_residuo_grid(Grid* g){
	int j,k;
	int nx = g->get_nx();
    int ny = g->get_ny();

    Matriz<double>& r = g->r;
    Matriz<double>& r_temp = g->r_temp;
    Matriz<double>& c = g->c;
    
    for (j=1 ; j < nx-1 ; j++){
		for (k=1 ; k < ny-1 ; k++){
			r_temp(j,k) = r(j,k) + beta*(c(j-1,k) + c(j+1,k)) + beta*alpha*(c(j,k+1) + c(j,k-1)) -c(j,k)*gama;
		}
	}
}
void Multigrid::Restricao_fw (Grid* coarse, Grid* fine){
	unsigned int i,j;
	unsigned int nx = coarse->get_nx();
	unsigned int ny = coarse->get_ny();
	
	Matriz<double>& r_fine = fine->r_temp;
		
	for(i=1 ; i < nx-1 ; i++){
		for(j=1 ; j < ny-1 ; j++){
			coarse->r(i,j) = (1.0e+00/16.0e+00)*(r_fine(2*i-1,2*j-1) + r_fine(2*i-1,2*j+1) + r_fine(2*i+1,2*j-1) + r_fine(2*i+1,2*j+1) +
											  2.0e+00*(r_fine(2*i,2*j-1) + r_fine(2*i,2*j+1) + r_fine(2*i-1,2*j) + r_fine(2*i+1,2*j)) +
											  4.0e+00*(r_fine(2*i,2*j)));
		}
	}
}
void Multigrid::Prolonga_corrige (Grid* coarse, Grid* fine){
	unsigned int i,j;
	unsigned int nx = coarse->get_nx();
	unsigned int ny = coarse->get_ny();
		
	Matriz<double>& c_fine = fine->c;
	
	for(i=0 ; i < nx -1 ; i++){
		for(j=0 ; j < ny -1 ; j++){
			c_fine(2*i,2*j) += coarse->c(i,j);
			c_fine(2*i+1,2*j) += (1.0e+00/2.0e+00)*(coarse->c(i,j) + coarse->c(i+1,j));
			c_fine(2*i,2*j+1) += (1.0e+00/2.0e+00)*(coarse->c(i,j) + coarse->c(i,j+1));
			c_fine(2*i+1,2*j+1) += (1.0e+00/4.0e+00)*(coarse->c(i,j) + coarse->c(i+1,j) + coarse->c(i,j+1) + coarse->c(i+1,j+1));
		}
	}	
}
void Multigrid::corrige_aproximacao(){
	int i,j;
	int nx = Grids[0]->get_nx();
	int ny = Grids[0]->get_ny();
	Matriz<double>& v = Grids[0]->v;
	Matriz<double>& c = Grids[0]->c;
	
	for(i=1 ; i < nx-1 ; i++){
		for(j=1 ; j < ny-1 ; j++){
			v(i,j) += c(i,j);
		}
	}
}

/* Metodos Publicos */
void Multigrid::inicializar_vetores(double tempo_inicial){
	int nx = Grids[0]->get_nx(),
		ny = Grids[0]->get_ny();
	int i,j;
	Matriz<par>& X = Grids[0]->X;
	Matriz<double>& v = Grids[0]->v;
	Matriz<double>& v_old = Grids[0]->v_old;
	//inicializando v_old e v
	for(i = 0 ; i < nx ; i++)
		for(j = 0 ; j < ny ; j++)
			v_old(i,j) = v(i,j) = funcao(X(i,j).x , X(i,j).y , tempo_inicial);
}
void Multigrid::V_ciclo(int v1, int v2 , double tempo){
	int i;
	int num_grids = Grids.size();

	//Relaxa em Av = f na malha mais fina
	Relaxar_Av(1 , tempo);
	
	//Calcula o residuo r=f-Av da malha mais fina
	calcular_residuo_fine(tempo);
	
	//Descendo para a malha mais grosseira
	for(i=0 ; i < num_grids-1 ; i++)
	{
		Grids[i]->c = 0; //zera o vetor de correcao
		Relaxar_Ac( Grids[i] , v1); //Relaxa em Ac = r com c = 0
		calcular_residuo_grid(Grids[i]);
		
		Restricao_fw(Grids[i+1], Grids[i]); //Passa o residuo de h -> 2h
	}

	Grids[num_grids-1]->c = 0; //zera vetor de correcao
	Relaxar_Ac( Grids[num_grids-1], 1); // Ac = r com c = 0

	//Subindo para malha mais fina
	for(i = num_grids-2 ; i >= 0 ; --i) 
	{
		Prolonga_corrige(Grids[i+1], Grids[i]); //Corrige ch += c2h (interpolado)
		Relaxar_Ac( Grids[i], v2); //Relaxa em Ac = r 
	}
	corrige_aproximacao();
}
void Multigrid::atualizar_cc(double t){
	int nx = Grids[0]->get_nx(),
		ny = Grids[0]->get_ny();
	int i,j;
	Matriz<double>& v = Grids[0]->v;
	Matriz<double>& v_old = Grids[0]->v_old;
	Matriz<par>& X = Grids[0]->X;
	
	for (i=1 ; i< nx-1 ; i++)
			for (j=1 ; j < ny-1 ; j++)
				v_old(i,j) = v(i,j);
	
	//atualiza condicao de contorno
	for(i=0 ; i < nx ; i++)
	{
		v(i,0) = funcao(X(i,0).x , X(i,0).y , t);
		v(i,ny-1) = funcao(X(i,ny-1).x , X(i,ny-1).y , t);
	}
	for(i=0 ; i<ny ; i++)
	{
		v(0,i) = funcao(X(0,i).x , X(0,i).y , t);
		v(nx-1,i) = funcao(X(nx-1,i).x , X(nx-1,i).y , t);
	}
}

double Multigrid::max_residuo(Grid* g){
	int nx = g->get_nx(),
		ny = g->get_ny();
	int i,j;
	Matriz<double>& r = g->r;
	
	double residuo=-1000;
	
	//Anda somente nos pontos internos
	for(i=1 ; i< nx-1 ; i++){
		for(j=1 ; j< ny-1 ; j++){
			//calcula o erro
			if(fabs(r(i,j)) > residuo){
				residuo = fabs(r(i,j));
			}
		}
	}
	return residuo;
}

double Multigrid::max_erro(double t){
	int nx = Grids[0]->get_nx(),
		ny = Grids[0]->get_ny();
	int i,j;
	Matriz<par>& X = Grids[0]->X;
	Matriz<double>& v = Grids[0]->v;
	
	double erro=0.0;
	
	//Anda somente nos pontos internos
	for(i=1 ; i< nx-1 ; i++){
		for(j=1 ; j< ny-1 ; j++){
			//calcula o erro
			if(fabs(funcao(X(i,j).x, X(i,j).y, t) - v(i,j)) > erro){
				erro = fabs(funcao(X(i,j).x, X(i,j).y, t) - v(i,j));
			}
		}
	}
	return erro;
}

