/*
 * Relaxacao em Av=b e Ac=r
 * Calculo de residuo e correcao
 * Ciclo-V
 * Atualizacao de condicoes de fronteira
 * Operadores Multigrid
 */

#ifndef MULTIGRID_H_
#define MULTIGRID_H_
#include <iostream>
#include <fstream>
#include <vector>
#include <math.h>
#include "grid.h"
using namespace std;


class Multigrid{
	private:
		double hxf, hyf; // hx e hy da malha mais fina
		double hx2, hy2; // hx^2 e hy^2 da malha mais fina
		double tempo_final, tempo_inicial;
		
		double alpha;
		double beta;
		double gama;
		double mi;
						
		void Relaxar_Av(unsigned int num_iteracoes, double t);
		void Relaxar_Ac(Grid* g, unsigned int num_iteracoes);
		void calcular_residuo_fine(double t);
		void calcular_residuo_grid(Grid* g);
		void Restricao_fw (Grid* coarse, Grid* fine);
		void Prolonga_corrige (Grid* coarse, Grid* fine);
		void corrige_aproximacao();
		
	public:
		vector<Grid*>& Grids;
		double ht;		
		vector<double> erro_maximo();
		
		Multigrid(vector<Grid*>& _Grids, double _ht, double ti, double tf) : Grids(_Grids){
			mi = 1.0e+00;
			tempo_inicial = ti;
			tempo_final = tf;
			hxf=_Grids[0]->get_hx();
			hyf=_Grids[0]->get_hy();
			hx2 = hxf*hxf;
			hy2 = hyf*hyf;
			ht = _ht;
			alpha = hx2/hy2;
			beta = mi*ht;
			gama = hx2 + 2.0*beta + 2.0*beta*alpha;;		
		}
		~Multigrid(){}
				
		void inicializar_vetores(double tempo_inicial);
		void V_ciclo(int v1, int v2 , double tempo);
		void atualizar_cc(double t);

		double max_residuo(Grid* g);
		double max_erro(double t);
};
#endif /*MULTIGRID_H_*/